#ifndef __SSCAUX_H_INCLUDED__
#define __SSCAUX_H_INCLUDED__
#include <Rcpp.h>
#include <RcppEigen.h>
#include <cmath>

// [[Rcpp::depends(RcppEigen)]]
using namespace Rcpp;
using namespace Eigen;

using Eigen::Map;
using Eigen::MatrixXd;
using Eigen::VectorXd;
using Eigen::Lower;

// sine function for MatrixXd-class
MatrixXd sin(MatrixXd matX){
  const int n = matX.rows();
  const int m = matX.cols();
  MatrixXd matY(n,m);
  for(int i=0; i<n; i++){
    for(int j=0; j<m; j++){
      matY(i,j)=sin(matX(i,j));  
  } }
  return matY;
}

// empirical quantile function
double quantile(VectorXd vecX, double alpha){
  const int n = vecX.size();
  NumericVector vecY(n);
  for(int t=0; t<n; t++){vecY(t)=vecX(t);}
  std::sort(vecY.begin(),vecY.end());
  return vecY[alpha*n-1];
}


// integer vector for resampling
IntegerVector boot(IntegerVector vecI, const int n){
  const int i_min=min(vecI);
  const int i_max=max(vecI);
  NumericVector vecSraw = runif(n,i_min,i_max);
  IntegerVector vecS(n);
  for(int i=0; i<n; i++){
    vecS[i]=vecSraw[i];
  }
  return(vecS);
}

// colwise center of matrix
VectorXd center(MatrixXd matZ){
  const int m = matZ.cols();
  const int n = matZ.rows();
  VectorXd vecMu(m);
  for(int i=0; i<m; i++){
    VectorXd vecZ = matZ.col(i);
    vecMu(i)=vecZ.sum()/n;
  }
  return vecMu;
}

// colwise sums of matrix
VectorXd col_sums(MatrixXd matZ){
  const int m = matZ.cols();
  VectorXd vecMu(m);
  for(int i=0; i<m; i++){
    VectorXd vecZ = matZ.col(i);
    vecMu(i)=vecZ.sum();
  }
  return vecMu;
}

// vector of stacked first cross-moments
VectorXd Mij_coef(MatrixXd matZ){
  const int n = matZ.rows();
  const int m = matZ.cols();
  const int lmax = 0.5*m*(m-1);
  VectorXd vecMij(lmax);
  VectorXd vecZij(n);
  int l = 0;
  for(int i=0; i<(m-1); i++){
    for(int j=i+1; j<m; j++){
      for(int t=0; t<n; t++){vecZij[t]=matZ(t,i)*matZ(t,j);}
      vecMij(l)=vecZij.sum()/n;
      l=l+1;
    }
  }
  return vecMij;
}

// matrix of stacked first cross-moments
MatrixXd Mij_mat(MatrixXd matZ){
  const int n = matZ.rows();
  const int m = matZ.cols();
  MatrixXd matMij(m,m);
  VectorXd vecZij(n);
  for(int i=0; i<m; i++){
    for(int j=0; j<m; j++){
      for(int t=0; t<n; t++){vecZij[t]=matZ(t,i)*matZ(t,j);}
      matMij(i,j)=vecZij.sum()/n;
    }
  }
  return matMij;
}
  
// vector of stacked pearsons correlation coefficient, matrix as input argument
VectorXd corr_coef(MatrixXd matZ){
  const int m = matZ.cols();
  const int n = matZ.rows();
  const int lmax = 0.5*m*(m-1);
  double mu[m];
  double sigma[m];
  for(int i=0; i<m; i++){
    VectorXd vecZi = matZ.col(i);
    mu[i]=vecZi.sum()/n;
    sigma[i]=pow(vecZi.array(),2.0).sum()/n-pow(mu[i],2);
  }
  VectorXd vecRho(lmax);
  VectorXd vecZij(n);
  int l=0;
  for(int i=0; i<(m-1); i++){
    for(int j=i+1; j<m; j++){
      for(int t=0; t<n; t++){vecZij[t]=matZ(t,i)*matZ(t,i);}
      vecRho(l)=(vecZij.sum()-n*mu[i]*mu[j])*pow(sigma[i],-0.5)*pow(sigma[j],-0.5)/n;
      l=l+1;
    }
  }
  return vecRho;
}

// pearsons correlation matrix
MatrixXd corr_mat(MatrixXd matZ){
  const int n = matZ.rows();
  const int m = matZ.cols();
  double mu[m];
  double sigma[m];
  for(int j=0; j<m; j++){
    VectorXd vecZj = matZ.col(j);
    for(int i=0; i<n; i++){ vecZj[i]=matZ(i,j);}
    mu[j]=vecZj.sum()/n;
    sigma[j]=pow(vecZj.array(),2).sum()/n-pow(mu[j],2);
    }
  MatrixXd matRho(m,m);
  VectorXd vecZjk(n);
  for(int j=0; j<m; j++){
    for(int k=0; k<m; k++){
      for(int i=0; i<n; i++){vecZjk[i]=matZ(i,j)*matZ(i,k);}
      if(j==k){ matRho(j,k)=1;}
      matRho(j,k)=(vecZjk.sum()-n*mu[j]*mu[k])*pow(sigma[j],-0.5)*pow(sigma[k],-0.5)/n;
    }
  }
  return matRho;
}

// covariance matrix
MatrixXd cov_mat(MatrixXd matZ){
  const int n = matZ.rows();
  const int m = matZ.cols();
  double mu[m];
  VectorXd vecZ(n);
  for(int j=0; j<m; j++){
    for(int i=0; i<n; i++){ vecZ(i)=matZ(i,j);}
    mu[j]=vecZ.sum()/n;
  }
  MatrixXd matSigma(m,m);
  for(int j=0; j<m; j++){
    for(int k=0; k<m; k++){
      for(int i=0; i<n; i++){vecZ(i)=matZ(i,j)*matZ(i,k);}
      matSigma(j,k)=vecZ.sum()/n-mu[j]*mu[k];
    }
  }
  return matSigma;
}

// asymptotic variance of stacked first cross-moments, compare Steiger-paper
MatrixXd Oij_asym(MatrixXd matZ){
  const int n = matZ.rows();
  const int m = matZ.cols();
  const int lmax = 0.5*m*(m-1);
  double part1;
  double part2;
  double part3;
  double part4;
  MatrixXd matO(lmax,lmax);
  MatrixXd matM = Mij_mat(matZ);
  int i = -1; // row in cov-matrix
  for(int i1=0; i1<(m-1); i1++){
    for(int i2=(i1+1); i2<m; i2++){
      i = i+1;
      int j = -1; // col in cov-matrix
      for(int j1=0; j1<(m-1); j1++){
        for(int j2=(j1+1); j2<m; j2++){
        j = j+1;
        if(j<i){matO(i,j)=matO(j,i);}
        else{
          part1=(matM(i1,j1)-matM(i1,i2)*matM(i2,j1))*(matM(i2,j2)-matM(i2,j1)*matM(j1,j2));
          part2=(matM(i1,j2)-matM(i1,j1)*matM(j1,j2))*(matM(i2,j1)-matM(i1,i2)*matM(i1,j1));
          part3=(matM(i1,j1)-matM(i1,j2)*matM(j1,j2))*(matM(i2,j2)-matM(i1,i2)*matM(i1,j2));
          part4=(matM(i1,j2)-matM(i1,i2)*matM(i2,j2))*(matM(i2,j1)-matM(i2,j2)*matM(j1,j2));
          matO(i,j)=0.5/n*(part1+part2+part3+part4);}
        }
      }
    }
  }
  return matO;
}

// bootstrapped variance of stacked first cross-moments
MatrixXd Oij_boot(MatrixXd matX, const int B){
  const int n = matX.rows();
  const int m = matX.cols();
  const int lmax = 0.5*m*(m-1);
  MatrixXd matX_boot(n,m);
  MatrixXd matM12_boot(B,lmax);
  VectorXd vecM12_r(lmax);
  IntegerVector vecInd(n);
  for(int t=0; t<n; t++){vecInd[t]=t;}
  // repeat B times
  for(int r=0; r<B; r++){
    IntegerVector vecInd_boot = boot(vecInd,n);
    for(int i=0; i<m; i++){
      for(int t=0; t<n; t++){ matX_boot(t,i) = matX(vecInd_boot[t],i); }}
    vecM12_r = Mij_coef(matX_boot);
    for(int l=0; l<lmax; l++){matM12_boot(r,l)=vecM12_r(l);}
  }
  MatrixXd matO12 = cov_mat(matM12_boot);
  return matO12;
}

// Kendalls tau matrix
MatrixXd tau_mat(MatrixXd matX){
  const int n = matX.rows();
  const int m = matX.cols();
  double val;
  double per = R::choose(n,2);
  MatrixXd matTau(m,m);
  for(int i=0; i<m; i++){
    matTau(i,i)=1;
    for(int j=(i+1); j<m; j++){
      // rank correlation for each bivariate margin
      int Tauij=0;
      for(int t1=0; t1<(n-1); t1++){
        for(int t2=(t1+1); t2<n; t2++){
          val = (matX(t1,i)-matX(t2,i))*(matX(t1,j)-matX(t2,j));
          if(val>0){Tauij = Tauij + 1;}
          else if(val<0){Tauij = Tauij - 1;}
          else{Tauij=Tauij;}
        }
      }
      matTau(i,j)=pow(per,-1)*Tauij;
      matTau(j,i)=pow(per,-1)*Tauij;
    }
  }
  return matTau;
}

// supremum of Euclidean norm of Brownian Bridge
// [[Rcpp::export]]
double supBrownianBridge095(double Omega){
  VectorXd supBBnorm(1000);
  NumericVector vecBM(10000);
  NumericVector vecBB(10000);
  NumericVector vecBBnorm(10000);
  NumericVector vecIncr(10000);
  for(int b=0; b<1000; b++){
    vecIncr = rnorm(10000,0,0.01);
    vecBM(0) = vecIncr(0);
    for(int t=1; t<10000; t++){vecBM(t)=vecIncr(t)+vecBM(t-1);}
    for(int t=0; t<10000; t++){vecBB(t)=vecBM(t)-vecBM(9999)*(t+1)/10000;}
    // Euclidean Norm
    for(int t=1500; t<8500; t++){
      vecBBnorm(t) = pow(pow(vecBB(t),2)*Omega,0.5);
    }
    supBBnorm(b) = max(vecBBnorm);
  }
  double supBBnorm095 = quantile(supBBnorm,0.95);
  return supBBnorm095;
}

// critical values for squared tied-down Bessel process over (0.15,0.85)
// [[Rcpp::export]]
double supBesselProcess095(double Omega){
  VectorXd supBessel(1000);
  NumericVector vecBM(10000);
  NumericVector vecBB(10000);
  NumericVector vecBessel(10000);
  NumericVector vecIncr(10000);
  for(int b=0; b<1000; b++){
    vecIncr = rnorm(10000,0,0.01);
    vecBM(0) = vecIncr(0);
    for(int t=1; t<10000; t++){vecBM(t)=vecIncr(t)+vecBM(t-1);}
    for(int t=0; t<10000; t++){vecBB(t)=vecBM(t)-vecBM(9999)*(t+1)/10000;}
    // Euclidean Norm
    for(int t=1500; t<8500; t++){
      vecBessel(t) = pow(vecBB(t),2)*Omega/(0.0001*t*(1-0.0001*t));
    }
    supBessel(b) = max(vecBessel);
  }
  double supBessel095 = quantile(supBessel,0.95);
  return supBessel095;
}

// supremum of Euclidean norm of Brownian Bridge
double supBrownianBridgeMV095(int m){
  VectorXd supBBnorm(1000);
  MatrixXd matBM(10000,m);
  MatrixXd matBB(10000,m);
  VectorXd vecBBnorm(10000);
  VectorXd vecIncr(m);
  for(int b=0; b<1000; b++){
    for(int i=0; i<m; i++){
      matBM(0,i) = R::rnorm(0,0.01);
      for(int t=1; t<10000; t++){matBM(t,i)=R::rnorm(0,0.01)+matBM(t-1,i);}
      for(int t=0; t<10000; t++){matBB(t,i)=matBM(t,i)-matBM(9999,i)*(t+1)/10000;}
    }
    // Euclidean Norm
    for(int t=0; t<10000; t++){
      vecIncr = matBB.row(t);
      vecBBnorm(t) = vecIncr.norm();
    }
    supBBnorm(b) = vecBBnorm.lpNorm<Infinity>();
  }
  double supBBnorm095 = quantile(supBBnorm,0.95);
  return supBBnorm095;
}


#endif
