#ifndef __TAUX_H_INCLUDED__
#define __TAUX_H_INCLUDED__
#include <Rcpp.h>
#include <RcppEigen.h>
#include <cmath>
#include "0_SeqStrucChange_aux.h"

// [[Rcpp::depends(RcppEigen)]]
using namespace Rcpp;
using namespace Eigen;

using Eigen::Map;
using Eigen::MatrixXd;
using Eigen::VectorXd;
using Eigen::Lower;

// function for line search over nu in multivariate t-distribution
double nu_fct1(VectorXd vecW, const double nu, const int m){
  const int n = vecW.rows();
  ArrayXd vecWtrans = (vecW.array()).log() - vecW.array();
  double y = vecWtrans.sum()/n - R::digamma(0.5*nu) + R::digamma(0.5*(nu+m)) -log(0.5*(nu+m)) + log(0.5*nu) + 1;
  return y;
}

// function for line search over nu in t-Copula
double nu_fct2(MatrixXd matU, MatrixXd matRho, double nu){
  const int m = matU.cols();
  const int n = matU.rows();
  MatrixXd matY(n,m);
  VectorXd vecLderiv(n);
  VectorXd vecY_obs(m);
  for(int i=0; i<m; i++){
    for(int t=0; t<n; t++){
      matY(t,i)=R::qt(matU(t,i),nu,true,false);}}
  const MatrixXd matRho_inv = matRho.inverse();
  double ll_const = 0.5*n*(R::digamma(0.5*(nu+m)) + (m-1)*R::digamma(0.5*nu) - m*R::digamma(0.5*(nu+1)));
  double delta; // Mahalanobis Distance
  double yit_sq;
  VectorXd a(m);
  VectorXd b(m);
  for(int t=0; t<n; t++){
    vecY_obs = matY.row(t);
    delta = vecY_obs.transpose() * matRho_inv * vecY_obs;
    for(int i=0; i<m; i++){
      yit_sq = pow(vecY_obs(i),2);
      a(i) = yit_sq/(nu+yit_sq);
      b(i) = log(1+yit_sq/nu);
      }
    vecLderiv(t) = 0.5*b.sum() - a.sum()*0.5*(nu+1)/nu + (0.5*(nu+m)/nu) * (delta/(nu+delta)) - 0.5*log(1+delta/nu);
  }
  double res = ll_const + vecLderiv.sum();
  return res;
}

// mle of t-distribution with fixed degrees of freedom using em-algorithm
List mle_t(MatrixXd matX, double nu=0, int niter = 1){
  const int n = matX.rows();
  const int m = matX.cols();
  // variables for e-step
  VectorXd vecX(n);
  VectorXd vecMu = center(matX); // initial value for mu
  MatrixXd matXi = cov_mat(matX); // initial value for
  MatrixXd matXi_inv(m,m);
  double nu_new;
  if(nu==0){
    vecX=matX.col(0);
    double mu4 = ((vecX.array()).pow(4)).sum()/n;
    double mu2 = ((vecX.array()).pow(2)).sum()/n;
    nu_new = 6/(mu4/pow(mu2,2)-3)+4;}
  else{nu_new=nu;}
  double Stau;
  VectorXd vecW(n); // weights
  VectorXd vecObs_cent(m); // subvector for one observation (centered)
  double delta; // Mahalanobis distance
  // variables for m-step
  double f_new;
  double nu_lower; 
  double nu_upper;
  double f_lower;
  double f_upper;
  // main loop
  for(int j=0; j<niter; j++)
  {
  // e-step: update according to equation 18
  matXi_inv = matXi.inverse();
  for(int t=0; t<n; t++){
    for(int j=0; j<m; j++){vecObs_cent(j) = matX(t,j)-vecMu(j);}
    delta = vecObs_cent.transpose() * matXi_inv * vecObs_cent;
    vecW(t) = (nu_new + m)/(nu_new + delta);}
  // m-step 1
  // update according to equation 19
  Stau = vecW.sum();
  // update according to equation 20
  for(int j=0; j<m; j++){ 
    vecX = matX.col(j);  
    MatrixXd mu_temp = vecW.transpose() * vecX / Stau;
    vecMu(j) = mu_temp(0,0) ;
  }
  // update according to equation 21
  MatrixXd matStauXX(m,m);
  for(int t=0; t<n; t++){
    for(int j=0; j<m; j++){vecObs_cent(j) = matX(t,j)-vecMu(j);}
    MatrixXd tauXX = vecObs_cent * vecObs_cent.transpose();
    matStauXX = matStauXX + vecW(t) * tauXX;
  }
  matXi = matStauXX / n;
  if(nu == 0){
  // reset variables for m-step 2
  nu_lower = 1;
  nu_upper = 100;
  f_lower = nu_fct1(vecW,nu_lower,m);
  f_upper = nu_fct1(vecW,nu_upper,m); 
  // m-step 2
  int counter = 0;
  while(fabs(nu_upper-nu_lower)>=0.01){
    if(f_upper*f_lower>0){
      nu_new=101;
      break;}
    counter = counter + 1;
    nu_new=(nu_lower+nu_upper)/2.0;
    f_new=nu_fct1(vecW,nu_new,m);
    // move section boundaries accordingly                            
    if(f_new==0){
      break;}
    else if(f_lower*f_new>0){
      nu_lower=nu_new;
      f_lower=f_new;}
    else{
      nu_upper=nu_new;
      f_upper=f_new;}
    // break if number of iterations is reached
    if(counter==100){break;}
  } }
}
  return List::create(
    Named("Mu")=vecMu,
    Named("Xi")=matXi,
    Named("nu")=nu_new);
}

// log-likelihood for i.i.d. sample from multivariate t-distribution
double logl_t(MatrixXd matX, VectorXd vecMu, MatrixXd matXi, double nu){
  const int n = matX.rows();
  const int m = matX.cols();
  double ll = n*(R::lgammafn(0.5*(nu+m)) - R::lgammafn(0.5*nu) - 0.5*m*log(nu*M_PI) - 0.5*log(matXi.determinant()));
  VectorXd vecObs_cent(m); 
  VectorXd ll_contrib(n);
  double delta; // Mahalanobis distance
  MatrixXd matXi_inv = matXi.inverse();
  for(int t=0; t<n; t++){
    for(int j=0; j<m; j++){vecObs_cent(j) = matX(t,j)-vecMu(j);}
    delta = vecObs_cent.transpose() * matXi_inv * vecObs_cent; 
    ll_contrib(t) = - 0.5 * (nu+m) * log(1 + delta / nu);
  }
  return ll + ll_contrib.sum();
}

// log-likelihood for i.i.d. sample from univariate t-distribution
double logl_t(VectorXd vecX, double mu, double xi, double nu){
  const int n = vecX.size();
  VectorXd ll_contrib(n);
  double ll;
  if(nu==101){ll=-0.5*log(2*M_PI)-0.5*n*log(xi)-0.5*n;}
  else{
  double delta; // Mahalanobis distance
  ll = n*(R::lgammafn(0.5*(nu+1)) - R::lgammafn(0.5*nu) - 0.5*log(nu*M_PI) - 0.5*log(xi));
  for(int t=0; t<n; t++){
    delta = pow(vecX(t)-mu,2)/xi; 
    ll_contrib(t) =  - 0.5 * (nu+1) * log(1 + delta / nu);}
  }
  return ll + ll_contrib.sum();
}

// log-likelihood for i.i.d. sample from t-Copula
double logl_tCopula(MatrixXd matU, MatrixXd matRho, double nu){
  const int n = matU.rows();
  const int m = matU.cols();
  MatrixXd matY(n,m);
  VectorXd vecY_obs(m);
  // constant part of ll
  for(int i=0; i<m; i++){for(int t=0; t<n; t++){matY(t,i)=R::qt(matU(t,i),nu,true,false);}}
  double ll = n*(R::lgammafn(0.5*(nu+m))+(m-1)*R::lgammafn(0.5*nu)-m*R::lgammafn(0.5*(nu+1))-0.5*log(matRho.determinant()));
  double delta; // Mahalanobis distance
  MatrixXd matRho_inv = matRho.inverse();
  for(int t=0; t<n; t++){
    vecY_obs = matY.row(t);
    double a=log(1+pow(vecY_obs(0),2)/nu);
    for(int i=1; i<m; i++){a=a+log(1+pow(vecY_obs(i),2)/nu);}
    delta = vecY_obs.transpose() * matRho_inv * vecY_obs; 
    ll = ll + 0.5*(nu+1)*a - 0.5*(nu+m)*log(1+delta/nu);
  }
  return ll;
}

// numerical derivative of log-likelihood
double nu_fct3(MatrixXd matU, MatrixXd matRho, double nu){
  double nu_low = nu-0.01;
  double nu_up = nu+0.01;
  double ld_low = logl_tCopula(matU,matRho,nu_low);
  double ld_up = logl_tCopula(matU,matRho,nu_up);
  double res = 50*(ld_up-ld_low);
  return res;
}

// pseudo-ML for t-Copula
List mle_tCopula(MatrixXd matU,double nu=0){
  MatrixXd matTau = tau_mat(matU);
  MatrixXd matRho = sin(0.5*M_PI*matTau);
  // initialize variables for bisection
  double nu_new;
  if(nu ==0){
  double f_new;
  double nu_lower = 1;
  double nu_upper = 20;
  double f_lower = nu_fct3(matU,matRho,nu_lower);
  double f_upper = nu_fct3(matU,matRho,nu_upper); 
  // m-step 2
  int counter = 0;
  while(fabs(nu_upper-nu_lower)>=0.01){
    if(f_upper*f_lower>0){
      nu_new=100;
      break;}
    counter = counter + 1;
    nu_new=(nu_lower+nu_upper)/2;
    f_new=nu_fct3(matU,matRho,nu_new);
    // move section boundaries accordingly                            
    if(f_new==0){
      break;}
    else if(f_lower*f_new>0){
      nu_lower=nu_new;
      f_lower=f_new;}
    else{
      nu_upper=nu_new;
      f_upper=f_new;}
    // break if number of iterations is reached
    if(counter==50){break;}
  }
  }
  else{
    nu_new=nu; 
  }
  return List::create(
    Named("Rho")=matRho,
    Named("nu")=nu_new);
}

#endif
